// Forlix SrcDS Manager
// http://forlix.org/, df@forlix.org
//
// Copyright (c) 2010-2013 Dominik Friedrichs

#include "SrcDS_Manager.h"

#define BELOW_NORMAL_PRIORITY_CLASS 0x00004000
#define ABOVE_NORMAL_PRIORITY_CLASS 0x00008000


bool CreateServerProcess(const MT_ServerData *mtsd,
                         PROCESS_INFORMATION *pi)
{
  char exec[MAX_PATH];
  sprintf(exec, "%s\\%s", mtsd->Path, GAME_EXEC);

  char args[MAX_PATH];
  sprintf(args, "arg0 -console"
                    " -game %s"
                    " -ip %s"
                    " -port %s"
                    " -maxplayers %s"
                    " +map %s"
                    " %s",
          mtsd->GameDir,
          mtsd->Host,
          mtsd->Port,
          mtsd->MaxPlayers,
          mtsd->MapName,
          mtsd->MiscArgs);

  STARTUPINFO si;
  memset(&si, 0, sizeof(si));
  si.cb = sizeof(STARTUPINFO);
  si.wShowWindow = SW_SHOWMINNOACTIVE;
  si.dwFlags = STARTF_USESHOWWINDOW;

  // determine priority
  DWORD priority;

  if(strequ(mtsd->Priority, "Low", false))
    priority = IDLE_PRIORITY_CLASS;
  else
  if(strequ(mtsd->Priority, "BelowNormal", false))
    priority = BELOW_NORMAL_PRIORITY_CLASS;
  else
  if(strequ(mtsd->Priority, "AboveNormal", false))
    priority = ABOVE_NORMAL_PRIORITY_CLASS;
  else
  if(strequ(mtsd->Priority, "High", false))
    priority = HIGH_PRIORITY_CLASS;
  else
  if(strequ(mtsd->Priority, "Realtime", false))
    priority = REALTIME_PRIORITY_CLASS;
  else
    priority = NORMAL_PRIORITY_CLASS;

  // launch server suspended
  if(!CreateProcess(exec, args, NULL, NULL, FALSE,
                    CREATE_SUSPENDED|priority, NULL, mtsd->Path, &si, pi))
    return(false);

  // set affinity
  DWORD sysaff;
  DWORD procaff;
  GetProcessAffinityMask(pi->hProcess, &procaff, &sysaff);

  DWORD affinity = strtoul(mtsd->Affinity, NULL, 16) & sysaff;
  SetProcessAffinityMask(pi->hProcess, affinity);

  ResumeThread(pi->hThread);
  return(true);
}


bool CreateUpdateProcess(const MT_ServerData *mtsd,
                         PROCESS_INFORMATION *pi,
                         HANDLE &pipeout_read,
                         bool toolonly)
{
  char exec[MAX_PATH];
  char workdir[MAX_PATH];

  sprintf(exec, "%s\\%s\\%s", mtsd->Path, mtsd->UpdPath, UPDATE_EXEC);
  GetFileDirectory(workdir, exec);

  char *args = NULL;

  if(toolonly)
    args = "arg0 +quit";

  HANDLE pipein_read;
  HANDLE pipein_write;

  HANDLE pipeout_write;

  SECURITY_ATTRIBUTES sa;
  memset(&sa, 0, sizeof(sa));

  sa.nLength = sizeof(SECURITY_ATTRIBUTES);
  sa.bInheritHandle = TRUE;
  sa.lpSecurityDescriptor = NULL;

  if(!CreatePipe(&pipeout_read, &pipeout_write, &sa, 2048)
  || !CreatePipe(&pipein_read, &pipein_write, &sa, 2048))
    return(false);

  STARTUPINFO si;
  memset(&si, 0, sizeof(si));
  si.cb = sizeof(STARTUPINFO);
  si.hStdInput = pipein_read;
  si.hStdOutput = pipeout_write;
  si.hStdError = pipeout_write;
  si.wShowWindow = SW_SHOWMINNOACTIVE;
  si.dwFlags = STARTF_USESHOWWINDOW|STARTF_USESTDHANDLES;

  if(!CreateProcess(exec, args, NULL, NULL, TRUE,
                    NORMAL_PRIORITY_CLASS|CREATE_NEW_CONSOLE, NULL, workdir, &si, pi))
  {
    CloseHandle(pipein_write);
    CloseHandle(pipein_read);

    CloseHandle(pipeout_read);
    CloseHandle(pipeout_write);
    return(false);
  }

  if(!toolonly)
  {
    char cmds[MAX_PATH+256];
    unsigned int cmds_len;

    cmds_len = sprintf(cmds, "login anonymous\n"
                             "force_install_dir \"%s\"\n"
                             "app_update %s\n"
                             "quit\n",
                             mtsd->Path, mtsd->UpdArgs);

    DWORD byteswritten;
    WriteFile(pipein_write, cmds, cmds_len, &byteswritten, NULL);
  }

  CloseHandle(pipein_write);
  CloseHandle(pipein_read);

  CloseHandle(pipeout_write);
  return(true);
}


ServerHealth GetServerHealth(const MT_ServerData *mtsd,
                             PROCESS_INFORMATION *pi,
                             unsigned int *pf)
{
  if(HasProcessTerminated(pi->hProcess))
  // server process terminated or crashed
    return(Server_Terminated);

  if(!PingSrcDS(mtsd->Host, mtsd->Port, 250))
  // ping attempt timed out
  // socket errors (-1) are not regarded as ping failure
  {
    if(++*pf >= 4)
    // admit timeout after 4 tries
      return(Server_TimedOut);
  }
  else
  // any successful ping resets the failure counter
    *pf = 0;

  PROCESS_MEMORY_COUNTERS pmem;
  unsigned long maxmem_mb = strtoul(mtsd->MaxMemory, NULL, 10);

  if(maxmem_mb
  && GetProcessMemoryInfo(pi->hProcess, &pmem, sizeof(pmem))
  && pmem.PagefileUsage/1024/1024 > maxmem_mb)
  // total memory usage (paged + nonpaged) exceeded the limit
    return(Server_MemoryFull);

  return(Server_Healthy);
}


void FetchFromPipe(HANDLE pipe, CStrBuf *buf)
{
  // peek if there are bytes to read
  DWORD bytesavail = 0;
  PeekNamedPipe(pipe, NULL, 0, NULL, &bytesavail, NULL);

  if(bytesavail)
  do
  {
    char tmp[256];

    // read only as much as the buffer can hold
    if(bytesavail > sizeof(tmp)-1)
      bytesavail = sizeof(tmp)-1;
  
    // we prevent ReadFile from blocking by telling it to
    // read less than or exactly as much bytes as available
    DWORD bytesread = 0;
    ReadFile(pipe, tmp, bytesavail, &bytesread, NULL);
    tmp[bytesread] = '\0';
    buf->Append(tmp);

    bytesavail = 0;
    PeekNamedPipe(pipe, NULL, 0, NULL, &bytesavail, NULL);
  }
  while(bytesavail);

  return;
}


void WriteUpdateLog(const MT_ServerData *mtsd,
                    const char *update_buf)
{
  FILE *fulog = fopen(UPDATELOG_PATH, "at");

  if(fulog)
  {
    time_t time_u = time(NULL);
    struct tm *lcltime = localtime(&time_u);

    fprintf(fulog, "L %04u/%02u/%02u - %02u:%02u:%02u: [UpdateLog for \"%s\" follows]\n",
                   lcltime->tm_year+1900, lcltime->tm_mon+1, lcltime->tm_mday,
                   lcltime->tm_hour, lcltime->tm_min, lcltime->tm_sec, mtsd->FriendlyName);

    fprintf(fulog, "%s\n", update_buf);
    fclose(fulog);
  }
  else
    LogMsg("[Update] Failed to write log to \"%s\"", UPDATELOG_PATH);

  return;
}
